#include "otp.h"

void
fatal(char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fprintf(stderr, "\n");
  exit(-1);
}

// called by pressing ctrl+c
void
close_key()
{
  close(key_fd);
  close(rand_fd);
  printf("\n[] key generated!\n[] Calculating the entropy of key...");
  fflush(stdout);
  printf("\r[] Entropy of the key is %.3f%%\n", entropia(arg.key_file));
  exit(0);
}


// print some needed informations
void
usage()
{
  printf("\n\n\topt implements the one time pad algo!\n\totp usage:\n\n\n");
  printf("\t- Create key:\n\t\totp --gen_key key_file\n\t\topt -g key_file\n\n");
  printf("\t- Create VPN with OTP:\n\t\totp -v server -r recv_key_file -s send_key_file\n\t\totp --vpn client server_host --recv_key recv_key_file --send_key send_key_file\n\n");
  printf("\t You can specify the offset of keys with -so (--send-offset) and -ro (--recv-offset)\n");
  printf("\tIMPORTANT NOTE: after setting up the vpn you can see the interface with 'ifconfig -a' but you must configure it\n");
  printf("\twith something like 'ifconfig otp0 192.168.0.1 pointopoint 192.168.0.2'.\n\tFor more verbose output you can use -D or --debug.\n\n\n");
  exit(-1);
}

// parse command lines argument
void
parse_args(int argc, char **argv)
{
#define GEN_KEY       1
#define VPN_CLIENT    2
#define VPN_SERVER    3
#define ENTROPY       4

  unsigned short int i;

  if (argc < 2)
    usage();

  arg.type = 0;
  // cycle through all arguments
  for (i = 1; i < argc; i++)
  {
    if (argv[i][0] == '-')
    {
      // debug
      if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--debug"))
        arg.debug = 1;
      // help
      else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
        usage();
      // genkey
      else if (!strcmp(argv[i], "-g") || !strcmp(argv[i], "--gen-key"))
      {
        if (arg.type)
          fatal("[] Incompatible option... '%s'\n", argv[i]);
		  
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument (file)\n", argv[i]);
        else
		{
          arg.key_file = argv[++i];
		  arg.type=GEN_KEY;
		}
      }
      //entropy
      else if (!strcmp(argv[1],"-e") || !strcmp(argv[i],"--entropy"))
      {
        if (arg.type)
          fatal("[] Incompatible option... '%s'\n", argv[i]);
		  
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument (file)\n", argv[i]);
        else
		{
          arg.key_file = argv[++i];
		  arg.type=ENTROPY;
		}
      }
      // vpn
      else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--vpn"))
      {
        if (arg.type)
          fatal("[] Incompatible option... '%s'\n", argv[i]);
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument (client/server)\n", argv[i]);
        i++;
        if (!strcmp(argv[i], "client"))
        {
          arg.type = VPN_CLIENT;
          if (argc - 1 <= i)
            fatal
              ("[] VPN tunnel in client mode needs the server (example --vpn client 213.121.123.213)\n");
          arg.remote_ip = argv[++i];
        } else if (!strcmp(argv[i], "server"))
          arg.type = VPN_SERVER;
        else
          fatal("[] '%s' option need an argument (client/server)\n", argv[i]);
      }
	  //recv-key
	  else if (!strcmp(argv[i],"-r") || !strcmp(argv[i],"--recv-key"))
	  {
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument\n", argv[i]);
        else
          arg.recv_file = argv[++i];
	  
	  }
	  //send-key
	  else if (!strcmp(argv[i],"-s") || !strcmp(argv[i],"--send-key"))
	  {
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument\n", argv[i]);
        else
          arg.send_file = argv[++i];
	  
	  }
	  //send offset
	  else if (!strcmp(argv[i],"-so") || !strcmp(argv[i],"--send-offset"))
	  {
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument\n", argv[i]);
        else
          arg.send_offset = atol(argv[++i]);
	  }
	  //recv offset
	  else if (!strcmp(argv[i],"-ro") || !strcmp(argv[i],"--recv-offset"))
	  {
        if (argc - 1 <= i)
          fatal("[] '%s' option need an argument\n", argv[i]);
        else
          arg.recv_offset = atol(argv[++i]);
	  }
	  
    } else
      fatal("[] Error while parsing arguments...what's '%s'?\n", argv[i]);
  }

  if (!arg.type)
    usage();

  
  // last control
  if (arg.type==VPN_CLIENT || arg.type==VPN_SERVER)
  {
	if (!arg.recv_file || !arg.send_file)
	  fatal("[] Missing need --send_key or --recv_key option\n");
  }
}


int
main(int argc, char **argv)
{

  // initialize variable
  arg.remote_ip = NULL;
  arg.recv_file = NULL;
  arg.send_file = NULL;
  arg.recv_offset=0;
  arg.send_offset=0;
  arg.debug = 0;


  // parse agruments
  parse_args(argc, argv);

  r_offset=arg.recv_offset;
  s_offset=arg.send_offset;


  if (arg.debug)
  {
    printf("[] After parse_args arg struct is: \n");
    printf("[] remote_ip : %s\n", arg.remote_ip);
    printf("[] type : %d\n", arg.type);
    printf("[] debug : %d\n", arg.debug);
	printf("[] r_offset : %lld\n",r_offset);
	printf("[] s_offset : %lld\n",s_offset);
  }

  switch (arg.type)
  {
	case ENTROPY:
	{
	  printf("[] Entropy of %s is %.3f%%\n",arg.key_file,entropia(arg.key_file));
	  break;
	}
    case GEN_KEY:
      {
        char key[1024];
        unsigned long long int l = 0;
        int i = 0;

        if ((key_fd = open(arg.key_file, O_CREAT | O_WRONLY, 00600)) < 0)
          fatal("Error while opening '%s' to write: '%s'\n", arg.key_file,
                strerror(errno));
        if ((rand_fd = open("/dev/urandom", O_RDONLY)) < 0)
          fatal("Error while opening '/dev/urandom' to read: '%s'\n",
                  strerror(errno));

        printf("[] Generating key....\n");
        printf("[] otp need more time to generate the random key.\n");
        printf("[] I hope you daemonize me with a ctrl+z and bg. \n");
        printf
          ("[] press CTRL+C when you decide that your key is of appropriate size!\n");

        signal(SIGINT, close_key);
        fflush(stdout);
        // take random data from /dev/random
        while ((i = read(rand_fd, key, 1024)))
        {
            printf("\r[]%lld bytes", l += i);
            write(key_fd, key, i);
            fflush(stdout);
        }
        fprintf(stderr, "[] Error while create key: '%s'\n", strerror(errno));
        close(rand_fd);
        close(key_fd);
        exit(-1);
      }
      break;


    //devo aprire 2 file di key e posizionarmi alla posizione corretta
    case VPN_CLIENT:
      {

		//open the key to decrypt data
        if ((recv_fd = open(arg.recv_file, O_RDONLY)) < 0)
          fatal("[] Error while open %s file: '%s'\n", arg.recv_file,
                strerror(errno));

		//going to the correct position
		if (r_offset)
		  if (lseek(recv_fd,r_offset,SEEK_SET)==-1)
			fatal("[] Error while going to correct position : '%s'\n",strerror(errno));

		//open the key to encrypt data
        if ((send_fd = open(arg.send_file, O_RDONLY)) < 0)
          fatal("[] Error while open %s file: '%s'\n", arg.send_file,
                strerror(errno));

		//going to the correct position
		if (s_offset)
		  if (lseek(send_fd,s_offset,SEEK_SET)==-1)
			fatal("[] Error while going to correct position : '%s'\n",strerror(errno));


		//open the tun device				
        vpn_fd = initTun();
		
		//open the tcp/ip connection (port 12000)
        if ((client_fd = otp_connect(arg.remote_ip, 12000)) < 0)
          fatal("[] Error while connection to '%s': %s\n", arg.remote_ip,
                strerror(errno));
				
		//uhm ??!?!?!?!?		
        // system("ifconfig otp0 10.0.0.2 pointopoint 10.0.0.1");
		
		//call engine
        main_loop();
        break;

      }

    case VPN_SERVER:
      {

		//open the key to decrypt data
        if ((recv_fd = open(arg.recv_file, O_RDONLY)) < 0)
          fatal("[] Error while open %s file: '%s'\n", arg.recv_file,
                strerror(errno));

		//going to the correct position
		if (r_offset)
		  if (lseek(recv_fd,r_offset,SEEK_SET)==-1)
			fatal("[] Error while going to correct position : '%s'\n",strerror(errno));

		//open the key to encrypt data
        if ((send_fd = open(arg.send_file, O_RDONLY)) < 0)
          fatal("[] Error while open %s file: '%s'\n", arg.send_file,
                strerror(errno));

		//going to the correct position
		if (s_offset)
		  if (lseek(send_fd,s_offset,SEEK_SET)==-1)
			fatal("[] Error while going to correct position : '%s'\n",strerror(errno));




		//open the tun device
        vpn_fd = initTun();
		
		//bind port (here i'm server)
        client_fd = otp_server(12000);

        // system("ifconfig otp0 10.0.0.1 pointopoint 10.0.0.2");
		
		//call engine
        main_loop();
		
        break;
      }
  }

  return 0;
}
